import os
import threading
import time

from sqlpharmacy.core import Database

from tek import ConfigClient, logger
from tek.config import lazy_configurable, Configurations
from tek.run import SignalManager

import series.config
import series.daemon.config
from series.daemon.model.release import ReleaseMonitor
from series.daemon.download import DownloadManager
from series.daemon.archiver import Archiver
from series.daemon.subsyncer import Subsyncer
from series.daemon.feed_poller import FeedPoller
from series.daemon.rest_api import RestApi
from series.daemon.errors import InvalidDBError

__all__ = ['SeriesDownloader']

conf = ConfigClient('series')


@lazy_configurable(series=['db_path'],
                   daemon=['run_downloader', 'run_archiver',
                           'run_subsyncer', 'run_feed_poller', 'run_rest_api'])
class SeriesDownloader(object):

    def __init__(self):
        self.__init_attributes()
        self._setup_db()
        self._setup_downloader()
        self._setup_archiver()
        self._setup_subsyncer()
        self._setup_feed_poller()
        self._setup_rest_api()

    def __init_attributes(self):
        self._need_commit = False
        self._commit_event = threading.Event()
        self._commit_event.set()
        self._queue = []

    def _setup_db(self):
        Database.register()
        _dir = os.path.dirname(self._db_path)
        if _dir and not os.path.isdir(_dir):
            os.makedirs(_dir)
        if os.path.isdir(self._db_path):
            raise InvalidDBError('Is a directory!')
        uri = 'sqlite:///{}'.format(self._db_path)
        self._db = Database(uri, connect_args=dict(check_same_thread=False))
        self._db.create_tables()
        self._session = self._db.session

    def _setup_downloader(self):
        self._downloader = DownloadManager(self._queue, self._wait_commit)

    def _setup_archiver(self):
        self._archiver = Archiver(self._queue, self._wait_commit)

    def _setup_subsyncer(self):
        self._subsyncer = Subsyncer(self._queue, self._wait_commit)

    def _setup_feed_poller(self):
        self._feed_poller = FeedPoller(self._session, self._queue, self._wait_commit)

    def _setup_rest_api(self):
        self._rest_api = RestApi(self._queue, self._wait_commit, self._session)

    def load_monitors(self):
        monitors = self._session.query(ReleaseMonitor).all()
        self._queue[:] = sorted(monitors, key=lambda m: m.index)
        logger.info('Loaded {} releases from db.'.format(len(self._queue)))

    def run(self):
        self.register_sigint_handler()
        self.load_monitors()
        if self._run_feed_poller:
            self._feed_poller.start()
        if self._run_downloader:
            self._downloader.start()
        if self._run_archiver:
            self._archiver.start()
        if self._run_subsyncer:
            self._subsyncer.start()
        if self._run_rest_api:
            self._rest_api.start()
        self.main_loop()

    def main_loop(self):
        self._running = True
        while self._running:
            time.sleep(5)
            self._commit()

    def _commit(self):
        if self._need_commit:
            self._commit_event.clear()
            self._session.commit()
            logger.debug('Committed transactions to db.')
            self._need_commit = False
            self._commit_event.set()

    def _wait_commit(self, block=True):
        self._need_commit = True
        if block:
            self._commit_event.wait()

    def register_sigint_handler(self):
        SignalManager.instance.sigint(self._interrupt)

    def _interrupt(self, signum, frame):
        self._running = False
        if self._run_feed_poller:
            self._feed_poller.stop()
        if self._run_downloader:
            self._downloader.stop()
        if self._run_archiver:
            self._archiver.stop()
        if self._run_subsyncer:
            self._subsyncer.stop()
        if self._run_rest_api:
            self._rest_api.stop()
        self._session.commit()
        if self._feed_poller.is_alive():
            self._feed_poller.join()
        if self._downloader.is_alive():
            self._downloader.join()
        if self._archiver.is_alive():
            self._archiver.join()
        if self._subsyncer.is_alive():
            self._subsyncer.join()
        if self._rest_api.is_alive():
            self._rest_api.join()


def daemon_cli():
    Configurations.parse_cli()
    SeriesDownloader().run()
