# -*- coding: utf-8 -*-


import os

from tek import logger
from tek.config import lazy_configurable, Configurations

from series import store_episode
from series.store_episode import EpisodeHandler as EpisodeStore
from series.daemon.errors import ArchiverError
from series.daemon.handler import Handler


@lazy_configurable(series=['series_dir'], daemon=['path_template'])
class Archiver(Handler):

    def __init__(self, queue, commit):
        super(Archiver, self).__init__(queue, commit, 5, 'archiver')
        Configurations.override('store_episode', overwrite=True)

    def _qualify(self, release):
        return (not (release.unknown or release.archived or
                     release.archiver_error)
                and release.downloaded)

    def _handle(self, monitor):
        release = monitor.release
        logger.info('Archiving {!s}.'.format(release))
        try:
            if not monitor.download_path:
                raise ArchiverError('Can\'t archive: no download path!')
            dest_path = self._setup_paths(monitor)
            self._store(monitor, dest_path)
        except store_episode.errors.UnknownSeries as e:
            logger.error(e)
            monitor.unknown = True
        except ArchiverError as e:
            logger.error(e)
            monitor.archiver_error = True
        self._commit()

    def _store(self, monitor, dest_path):
        handler = EpisodeStore(ask=False, series_name=monitor.release.name)
        handler.add_job(monitor.download_path, dest_path)
        if handler.store():
            monitor.archived = True

    def _setup_paths(self, monitor):
        local_path = self._format_path(monitor)
        dest_path = os.path.join(self._series_dir, local_path)
        target_dir = os.path.dirname(dest_path)
        self._check_target_dir(target_dir)
        self._check_downloaded_file(monitor, dest_path)
        return dest_path

    def _check_downloaded_file(self, monitor, dest_path):
        if not os.path.isfile(monitor.download_path):
            text = 'Downloaded release "{!s}" missing!'
            if os.path.isfile(dest_path):
                text += '\nHowever, it seems to be archived.'
                monitor.archived = True
            raise ArchiverError(text.format(monitor.release))

    def _check_target_dir(self, target_dir):
        if not os.path.exists(target_dir):
            os.makedirs(target_dir)
        elif not os.path.isdir(target_dir):
            raise ArchiverError('Not a directory: {}'.format(target_dir))

    def _format_path(self, monitor):
        parts = monitor.download_path.rsplit('.', 1)
        release = monitor.release
        if len(parts) <= 1:
            logger.error('Invalid filename. Assuming \'mkv\' extension.')
            extension = 'mkv'
        else:
            extension = parts[1]
        parameters = dict(
            name=release.name,
            season=release.season,
            episode=release.episode,
            ext=extension,
        )
        local = self._path_template.format(**parameters)
        return os.path.join(self._series_dir, local)
