from sqlpharmacy.core import Database

from sqlalchemy import Column, String, Integer, Boolean

from tek import logger, ConfigClient
from tek.tools import camelcaseify, find

from series import episode_enumeration, make_series_name
from series.daemon.model.link import Link


class Release(object, metaclass=Database.DefaultMeta):
    title = Column(String)
    name = Column(String)
    codec = Column(String)
    group = Column(String)
    season = Column(Integer)
    episode = Column(Integer)
    is_series = Column(Boolean)
    resolution = Column(String)

    @property
    def is_hd(self):
        return self.resolution in ['720p', '1080p']

    def __repr__(self):
        from tek.tools import repr_params
        params = repr_params(self.title, self.name, self.codec, self.group,
                             (self.season, self.episode), self.is_series,
                             self.resolution)
        return '{}{}'.format(self.__class__.__name__, params)

    def __str__(self):
        return '{} {}x{}'.format(camelcaseify(self.name, sep=' '), self.season,
                                 self.episode)

    def __eq__(self, other):
        return isinstance(other, Release) and self.title == other.title

    def is_same_episode(self, other):
        return (isinstance(other, Release) and self.name == other.name and
                self.season == other.season and self.episode == other.episode)

    @property
    def search_string(self):
        sep = '[._ ]'
        fix = '({}[^._ ]+)?'.format(sep)
        name = sep.join(self.name.split('_'))
        enum_format = lambda n: ('0?{}' if n < 10 else '{}').format(n)
        enum = '{}s?{}e?{}'.format(sep, enum_format(self.season),
                                   enum_format(self.episode))
        res = '{}{}'.format(sep, self.resolution) if self.resolution else ''
        return '{}{}{}{}'.format(name, enum, fix, res)


@Database.foreign_key(Release)
@Database.foreign_key('ReleaseMonitor', ref_name='_fix', backref_name='_nuke')
class ReleaseMonitor(object, metaclass=Database.DefaultMeta):
    index = Column(Integer)
    downloaded = Column(Boolean)
    download_path = Column(String)
    archived = Column(Boolean)
    _nuked = Column(Boolean)
    _links = Column(String)
    watched = Column(Boolean)
    unknown = Column(Boolean)
    down = Column(Boolean)
    archiver_error = Column(Boolean)
    failed_downloads = Column(Integer)
    downloading = Column(Boolean)
    subtitles_downloaded = Column(Boolean)

    def __init__(self, release, index, links):
        self.release = release
        self.index = index
        self._links = '|$|'.join(links)
        self.downloaded = False
        self.download_path = ''
        self.archived = False
        self._nuked = False
        self.unknown = False
        self.down = False
        self.archiver_error = False
        self.failed_downloads = 0
        self.downloading = False
        self.subtitles_downloaded = False
        self.subs_missing = False
        self._all_links_dead = False
        self._new_links = True

    def nuke(self, fix):
        self._nuked = True
        self._fix = fix

    @property
    def nuked(self):
        return self._nuked

    @property
    def links(self):
        if not hasattr(self, '_link_data'):
            self._link_data = [Link(link) for link in self._links.split('|$|')]
        return self._link_data

    @property
    def link(self):
        good = lambda links: find(lambda l: l.good, links)
        prefer = ConfigClient('daemon')('prefer_hosters')
        preferred = filter(lambda l: l.domain in prefer, self.links)
        preferred_link = good(preferred)
        return preferred_link or good(self.links)

    @property
    def explain_downloadable(self):
        text = '{:s} is {}downloadable{}.'
        if self.downloadable:
            nope = ''
            suffix = ''
        else:
            nope = 'not '
            suffix = ' because it {}'

            def list_reasons():
                status = '\n\t'.join(self._link_status)
                _reasons = (
                    ('is already downloaded', self.downloaded),
                    ('is nuked', self.nuked),
                    ('is down', self.down),
                    ('is a download in progress', self.downloading),
                    ('has no valid links:\n\t{}'.format(status),
                     self.all_links_dead),
                )
                for message, condition in _reasons:
                    if condition:
                        yield message
            reasons = list(list_reasons())
            if len(reasons) > 1:
                reasons[-2:] = [' and '.join(reasons[-2:])]
            reason = ', '.join(reasons)
            suffix = suffix.format(reason)
        return text.format(self.release, nope, suffix)

    @property
    def downloadable(self):
        return not (self.downloaded or self.nuked or self.down or
                    self.downloading or self.all_links_dead)

    def add_link(self, url):
        self.links.append(Link(url))
        if self._links:
            self._links += '|$|'
        self._links += url

    @property
    def all_links_dead(self):
        return all(link.invalid for link in self.links)

    @property
    def is_torrent(self):
        return self.link.is_torrent

    def add_links_from_feed_entry(self, feed_entry):
        my_links = self.links
        for link in feed_entry.links:
            if link not in my_links:
                self.add_link(link)
        added = set(self.links) - set(my_links)
        if added:
            text = 'Added new links to release {}: {}'
            logger.info(text.format(self.release, ', '.join(added)))

    @property
    def _link_status(self):
        return [l.status for l in self.links]


class ReleaseFactory(object):

    def __call__(self, title, name, codec, group, enum=None, res=None):
        is_series = bool(enum)
        enum = episode_enumeration(enum)
        name = make_series_name(name)
        codec = codec or ''
        group = group or ''
        return Release(title=title, name=name, codec=codec.lower(),
                       group=group.lower(), season=enum[0], episode=enum[1],
                       is_series=is_series, resolution=res)

    def monitor(self, entry, index):
        entry.resolve_redirects()
        return ReleaseMonitor(entry.release, index, entry.links)
