import threading
from functools import wraps

import requests

from flask import Flask, jsonify, request, json

from tek import logger
from tek.config import lazy_configurable

from series.daemon.releases_facade import ReleasesFacade


@lazy_configurable(daemon=['rest_api_host', 'rest_api_port'])
class RestApi(threading.Thread):

    def json_response(func):
        @wraps(func)
        def wrapper(*a, **kw):
            return jsonify(dict(response=func(*a, **kw)))
        return wrapper

    def __init__(self, queue, commit, session):
        self._releases = ReleasesFacade(queue, session)
        self._commit = commit
        threading.Thread.__init__(self)
        self.app = Flask('seriesd')

    def run(self):
        logger.info('Starting REST API.')
        self.setup_routes()
        self.app.run(host=self._rest_api_host, port=self._rest_api_port)

    def setup_routes(self):
        self.app.add_url_rule('/release/explain', 'explain', self._explain)
        self.app.add_url_rule('/release', 'release', self._release_list)
        self.app.add_url_rule('/release_id', 'release_id',
                              self._id_by_metadata)
        self.app.add_url_rule('/release/<int:_id>/link', 'add_link',
                              self._add_link, methods=['PUT'])
        self.app.add_url_rule('/release/<int:_id>/downloaded',
                              'mark_downloaded_id', self._mark_downloaded_id,
                              methods=['PUT'])
        self.app.add_url_rule('/release/<series>/<season>/<episode>/'
                              'downloaded',
                              'mark_downloaded_metadata',
                              self._mark_downloaded_metadata, methods=['PUT'])
        self.app.add_url_rule('/release/<series>/<season>/<episode>',
                              'create_release',
                              self._create_release, methods=['POST'])
        self.app.add_url_rule('/download/pending', 'pending_downloads',
                              self._pending_downloads)

    def _shutdown(self):
        func = request.environ.get('werkzeug.server.shutdown')
        func()
        return 'server shut down.'

    def stop(self):
        self.app.add_url_rule('/shutdown', 'shutdown', self._shutdown,
                              methods=['POST'])
        uri = 'http://localhost:{}/shutdown'.format(self._rest_api_port)
        requests.post(uri)

    @json_response
    def _explain(self):
        return [r.explain_downloadable for r in self._releases[-5:]]

    @json_response
    def _add_link(self, _id):
        data = json.loads(request.data)
        url = data['url']
        release = self._releases.find_by_id(_id)
        text = 'RestAPI: Adding link "{}" to release {}'
        logger.info(text.format(url, release.release))
        release.add_link(url)
        self._commit(block=False)

    @json_response
    def _pending_downloads(self):
        releases = self._releases.pending_downloads
        dls = [(r.id, r.release.search_string) for r in releases]
        return dls

    @json_response
    def _mark_downloaded_metadata(self, series, season, episode):
        monitor = self._releases.find_by_metadata(series, season, episode)
        return self._mark_downloaded(monitor)

    @json_response
    def _mark_downloaded_id(self, _id):
        monitor = self._releases.find_by_id(_id)
        return self._mark_downloaded(monitor)

    @json_response
    def _release_list(self):
        data = json.loads(request.data) if request.data else {}
        regex = data.get('regex', '')
        matches = self._releases.filter_episode_repr(regex)
        return [[str(monitor.id), str(monitor.release)] for monitor in matches]

    @json_response
    def _id_by_metadata(self):
        data = json.loads(request.data)
        results = self._releases.find_by_metadata(data['series'],
                                                  data['season'],
                                                  data['episode'])
        return results.id if results else -1

    @json_response
    def _create_release(self, series, season, episode):
        if self._releases.find_by_metadata(series, season, episode):
            return 'Release already exists'
        else:
            self._releases.create(series, season, episode)
            return 'ok'

    def _mark_downloaded(self, monitor):
        if monitor:
            msg = 'Marking release "{}" as downloaded.'
            logger.info(msg.format(monitor.release))
            monitor.downloaded = True
            self._commit(block=False)
        return monitor is not None

__all__ = ['RestApi']
