# coding: utf-8

import itertools
from datetime import datetime, timedelta

from tek.config import lazy_configurable
from tek.tools import first_valid, datetime_to_unix
from tek import logger

from series import EpisodeMetadataFactory, is_episode

from series.library.model.collection import Collection
from series.library.model.series import SeriesFactory, Series
from series.library.model.season import SeasonFactory, Season
from series.library.model.episode import EpisodeFactory, Episode
from series.library.model.watch_event import WatchEventFactory, WatchEvent


@lazy_configurable(library=['collection_paths'])
class LibraryFacade(object):

    def __init__(self, db_):
        self._db = db_
        self._epimeta_factory = EpisodeMetadataFactory()
        self._collections = [Collection(path) for path in
                             self._collection_paths]

    def scan(self):
        ''' Scan all present collections for new episodes.
        '''
        logger.info('Rescanning all collections…')
        files = itertools.chain(*[c.video_files for c in self._collections])
        new_files = 0
        for _file in files:
            if is_episode(_file):
                data = self._epimeta_factory.from_filename(_file)
                new_files += not self._episode_exists(data)
                self._add_new_episode(data)
        logger.info('Done. Found {} new episodes.'.format(new_files))

    def _episode_exists(self, data):
        return self.episode(series=data.series, season=data.season,
                            number=data.episode) is not None

    def _add_new_episode(self, data):
        ''' Add an episode with corresponding series and season objects
        if they don't exist yet.
        '''
        series = self._create_if_missing(SeriesFactory, name=data.series)
        season = self._create_if_missing(SeasonFactory, number=data.season,
                                         series=series)
        self._create_if_missing(EpisodeFactory, number=data.episode,
                                series=series, season=season)
        self._db.commit()

    def _create_if_missing(self, Factory, **params):
        ''' Call the DB factory function corresponding to _type with
        params, if there is no such object in the db yet.
        '''
        query = self._db.query(Factory.FACTORY_FOR).filter_by(**params)
        return query.first() if query.count() > 0 else Factory(**params)

    def series(self, name):
        ''' Return the series with the specified name.
        '''
        return self._db.query(Series).filter_by(name=name).first()

    def episodes(self, series=None, season=None, numbers=None):
        ''' Query the db for episodes, return all matches.
        If any of the arguments are specified, they are used to filter
        the results by series, season and episode number.
        '''
        query = self._db.query(Episode)
        if series is not None:
            if not isinstance(series, Series):
                series = self.series(series)
            query = query.filter_by(series=series)
        if season is not None:
            if not isinstance(season, Season):
                season = self.season(series, season)
            query = query.filter_by(season=season)
        if isinstance(numbers, list):
            query = query.filter(Episode.number.in_(numbers))
        return sorted(query.all())

    def episode(self, series, season, number):
        return first_valid(self.episodes(series=series, season=season,
                                         numbers=[number]))

    def seasons(self, series=None, numbers=None):
        query = self._db.query(Season)
        if series is not None:
            if not isinstance(series, Series):
                series = self.series(series)
            query = query.filter_by(series=series)
        if isinstance(numbers, list):
            query = query.filter(Season.number.in_(numbers))
        return sorted(query.all())

    def season(self, series, number):
        return first_valid(self.seasons(series=series, numbers=[number]))

    def episode_path(self, episode):
        results = [c.episode_path(episode) for c in self._collections]
        return first_valid(results)

    @property
    def new_episodes(self, *a, **kw):
        return filter(lambda e: e.new, self.episodes(*a, **kw))

    @property
    def all_series(self):
        return self._db.query(Series).all()

    def add_watch_event(self, series, season, number, start_time, finish_time):
        episode = self.episode(series, season, number)
        event = WatchEventFactory(time_begin=datetime_to_unix(start_time),
                                  time_end=datetime_to_unix(finish_time),
                                  episode=episode)
        self._db.session.add_then_commit(event)
        return event

    def recently_watched_episodes(self, days):
        thresh = datetime_to_unix(datetime.now() - timedelta(days=days))
        events = self._db.query(WatchEvent).filter(WatchEvent.time_begin >
                                                   thresh)
        episodes = [event.episode for event in events]
        return list(set(episodes))


__all__ = ['LibraryFacade']
